import typing

from starsessions.stores.base import SessionStore


class InMemoryStore(SessionStore):
    """Stores session data in a dictionary."""

    def __init__(self) -> None:
        self.data: typing.Dict[str, bytes] = {}

    async def read(self, session_id: str, lifetime: int) -> bytes:
        return self.data.get(session_id, b"")

    async def write(self, session_id: str, data: bytes, lifetime: int, ttl: int) -> str:
        self.data[session_id] = data
        return session_id

    async def remove(self, session_id: str) -> None:
        try:
            del self.data[session_id]
        except KeyError:
            pass

    async def exists(self, session_id: str) -> bool:
        return session_id in self.data



class SessionError(Exception):
    """Base class for session exceptions."""


class SessionNotLoaded(SessionError):
    """Raised on attempt to access/mutate session data of a session which has not been loaded yet."""

    solution = 'Call "starsessions.load_session(connection)" to load session data.'


class ImproperlyConfigured(SessionError):
    """Exception is raised when some settings are missing or misconfigured."""



from __future__ import annotations

import secrets
import time
import typing
from starlette.requests import HTTPConnection

from starsessions.exceptions import SessionNotLoaded
from starsessions.serializers import Serializer
from starsessions.stores import SessionStore
from starsessions.types import SessionMetadata


def generate_session_id() -> str:
    """Generate a new, cryptographically strong session ID."""
    return secrets.token_hex(16)


def regenerate_session_id(connection: HTTPConnection) -> str:
    """
    Generate new session ID and set it as current session ID.

    Returns new session ID.
    """
    return get_session_handler(connection).regenerate_id()


def get_session_id(connection: HTTPConnection) -> typing.Optional[str]:
    """Get current session ID."""
    return get_session_handler(connection).session_id


def get_session_handler(connection: HTTPConnection) -> SessionHandler:
    """
    Get session handler.

    Session handler is a tool for low level session management.
    """
    return typing.cast(SessionHandler, connection.scope["session_handler"])


async def load_session(connection: HTTPConnection) -> None:
    """
    Initialize session.

    Will replace any existing session data. Should be called once per request.
    """

    await get_session_handler(connection).load()


def is_loaded(connection: HTTPConnection) -> bool:
    """Test if session has been loaded for this connection."""
    return get_session_handler(connection).is_loaded


def get_session_metadata(connection: HTTPConnection) -> SessionMetadata:
    """
    Get session metadata. The session must be loaded first otherwise it raises.

    :raise SessionNotLoaded: if session is not loaded.
    """
    if not is_loaded(connection):
        raise SessionNotLoaded("Cannot read session metadata because session was not loaded.")

    metadata = get_session_handler(connection).metadata
    assert metadata  # satisfy mypy
    return metadata


def get_session_remaining_seconds(connection: HTTPConnection) -> int:
    """Get total seconds remaining before this session expires."""
    now = time.time()
    metadata = get_session_metadata(connection)
    return int((metadata["created"] + metadata["lifetime"]) - now)


class SessionHandler:
    """
    A tool for low level session management.

    This is private API, no backward compatibility guarantee.
    """

    def __init__(
        self,
        connection: HTTPConnection,
        session_id: typing.Optional[str],
        store: SessionStore,
        serializer: Serializer,
        lifetime: int,
    ) -> None:
        self.connection = connection
        self.session_id = session_id
        self.store = store
        self.serializer = serializer
        self.is_loaded = False
        self.initially_empty = False
        self.lifetime = lifetime
        self.metadata: typing.Optional[SessionMetadata] = None

    async def load(self) -> None:
        # don't refresh existing session, it may contain user data
        if self.is_loaded:  # pragma: nocover, IDK how to test it :(
            return

        self.is_loaded = True
        data = {}
        if self.session_id:
            data = self.serializer.deserialize(
                await self.store.read(
                    session_id=self.session_id,
                    lifetime=self.lifetime,
                ),
            )

        # read and merge metadata
        metadata = {"lifetime": self.lifetime, "created": time.time(), "last_access": time.time()}
        metadata.update(data.pop("__metadata__", {}))
        metadata.update({"last_access": time.time()})  # force update
        self.metadata = metadata  # type: ignore[assignment]

        self.connection.scope["session"] = {}
        self.connection.session.update(data)

        self.initially_empty = len(self.connection.session) == 0

    async def save(self, remaining_time: int) -> str:
        self.connection.session.update({"__metadata__": self.metadata})

        self.session_id = await self.store.write(
            session_id=self.session_id or generate_session_id(),
            data=self.serializer.serialize(self.connection.session),
            lifetime=self.lifetime,
            ttl=remaining_time,
        )
        return self.session_id

    async def destroy(self) -> None:
        """Destroy session."""
        if self.session_id:
            await self.store.remove(self.session_id)

    @property
    def is_empty(self) -> bool:
        return len(self.connection.session) == 0

    def regenerate_id(self) -> str:
        self.session_id = generate_session_id()
        return self.session_id



import functools
import typing
from redis.asyncio.client import Redis
from redis.asyncio.utils import from_url

from starsessions.exceptions import ImproperlyConfigured
from starsessions.stores.base import SessionStore


def prefix_factory(prefix: str, key: str) -> str:
    return prefix + key


if typing.TYPE_CHECKING:  # pragma: nocover
    BaseRedis = Redis[bytes]
else:
    BaseRedis = Redis


class RedisStore(SessionStore):
    """Stores session data in a Redis server."""

    def __init__(
        self,
        url: typing.Optional[str] = None,
        connection: typing.Optional[BaseRedis] = None,
        prefix: typing.Union[typing.Callable[[str], str], str] = "starsessions.",
        gc_ttl: int = 3600 * 24 * 30,
    ) -> None:
        """
        Initializes Redis session store. Either `url` or `connection` required. To namespace keys in Redis use `prefix`
        argument. It can be a string or callable that accepts a single string argument and returns new Redis key as
        string.

        :param url:  Redis URL. Defaults to None.
        :param connection: aioredis connection. Defaults to None
        :param prefix: Redis key name prefix or factory.
        :param gc_ttl: TTL for sessions that have no expiration time
        """
        if not (url or connection):
            raise ImproperlyConfigured("Either 'url' or 'connection' arguments must be provided.")

        if isinstance(prefix, str):
            prefix = functools.partial(prefix_factory, prefix)

        self.gc_ttl = gc_ttl
        self.prefix: typing.Callable[[str], str] = prefix
        if connection:
            self._connection: BaseRedis = connection
        else:
            assert url
            self._connection = from_url(url)

    async def read(self, session_id: str, lifetime: int) -> bytes:
        value = await self._connection.get(self.prefix(session_id))
        if value is None:
            return b""
        return value

    async def write(self, session_id: str, data: bytes, lifetime: int, ttl: int) -> str:
        if lifetime == 0:
            # Redis will fail for session-only cookies, as zero is not a valid expiry value.
            # We cannot know the final session duration so set here something close to reality.
            # FIXME: we want something better here
            ttl = self.gc_ttl

        ttl = max(1, ttl)
        await self._connection.set(self.prefix(session_id), data, ex=ttl)
        return session_id

    async def remove(self, session_id: str) -> None:
        await self._connection.delete(self.prefix(session_id))

    async def exists(self, session_id: str) -> bool:
        result: int = await self._connection.exists(self.prefix(session_id))
        return result > 0



from .exceptions import ImproperlyConfigured, SessionError, SessionNotLoaded
from .middleware import SessionAutoloadMiddleware, SessionMiddleware
from .serializers import JsonSerializer, Serializer
from .session import (
    generate_session_id,
    get_session_handler,
    get_session_id,
    get_session_metadata,
    get_session_remaining_seconds,
    is_loaded,
    load_session,
    regenerate_session_id,
)
from .stores import CookieStore, InMemoryStore, SessionStore

__all__ = [
    "SessionMiddleware",
    "SessionAutoloadMiddleware",
    "Serializer",
    "JsonSerializer",
    "SessionStore",
    "InMemoryStore",
    "CookieStore",
    "SessionError",
    "SessionNotLoaded",
    "ImproperlyConfigured",
    "get_session_id",
    "generate_session_id",
    "get_session_handler",
    "regenerate_session_id",
    "is_loaded",
    "load_session",
    "get_session_metadata",
    "get_session_remaining_seconds",
]



import typing
from base64 import b64decode, b64encode
from itsdangerous import BadSignature, TimestampSigner
from starlette.datastructures import Secret

from starsessions.stores.base import SessionStore


class CookieStore(SessionStore):
    """Stores session data in the browser's cookie as a signed string."""

    def __init__(self, secret_key: typing.Union[str, Secret]):
        self._signer = TimestampSigner(str(secret_key))

    async def read(self, session_id: str, lifetime: int) -> bytes:
        """A session_id is a signed session value."""
        try:
            if lifetime == 0:
                # For session-only cookies, the lifetime parameter has no sense
                # as the cookie will be deleted when browser closes.
                # but unsigning will always fail because the signature will always be expired.
                # So fake lifetime just for TimestampSigner success.
                # If you have a better solution please file an issue.
                lifetime = 3600 * 24 * 30 * 12  # 1 year
            data = self._signer.unsign(session_id, max_age=lifetime)
            return b64decode(data)
        except BadSignature:
            return b""

    async def write(self, session_id: str, data: bytes, lifetime: int, ttl: int) -> str:
        """The data is a session id in this storage."""
        encoded_data = b64encode(data)
        return self._signer.sign(encoded_data).decode("utf-8")

    async def remove(self, session_id: str) -> None:
        """Session data stored on client side - no way to remove it."""

    async def exists(self, session_id: str) -> bool:
        return False



import abc


class SessionStore(abc.ABC):  # pragma: no cover
    """Base class for session storages."""

    @abc.abstractmethod
    async def read(self, session_id: str, lifetime: int) -> bytes:
        """
        Read session data from the storage.

        :param session_id: ID associated with session
        :param lifetime: session lifetime duration
        :returns bytes: session data as bytes
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def write(self, session_id: str, data: bytes, lifetime: int, ttl: int) -> str:
        """
        Write session data to the storage.

        Must return session ID. Either new or from arguments.
        The difference between `lifetime` and `ttl` is that former defines the total session duration
        and the latter ony tells the amount to seconds left before the data can be safely removed.

        :param session_id: ID associated with session
        :param data: session data serialized to bytes
        :param lifetime: session lifetime, in seconds
        :param ttl: keep session data this amount of time, in seconds
        :returns str: session ID
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def remove(self, session_id: str) -> None:
        """
        Remove session data from the storage.

        :param session_id: ID associated with session
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def exists(self, session_id: str) -> bool:
        """
        Test if storage contains session data for a given session_id.

        :param session_id: ID associated with session
        :returns bool: True if there is data in session store otherwise False.
        """
        raise NotImplementedError()



import datetime
import re
import typing
from starlette.datastructures import MutableHeaders
from starlette.requests import HTTPConnection
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from starsessions import SessionNotLoaded
from starsessions.serializers import JsonSerializer, Serializer
from starsessions.session import SessionHandler, get_session_remaining_seconds, load_session
from starsessions.stores import SessionStore


class LoadGuard:
    """A guard that protects access to uninitialized session data."""

    def __getattribute__(self, item: str) -> typing.Any:
        if item == "_raise":
            return super().__getattribute__(item)

        self._raise()

    def __setitem__(self, key: str, value: typing.Any) -> typing.NoReturn:  # pragma: nocover
        self._raise()

    def __getitem__(self, key: str) -> typing.NoReturn:  # pragma: nocover
        self._raise()

    def _raise(self) -> typing.NoReturn:
        raise SessionNotLoaded("Attempt to access session that has not been loaded yet.")


class SessionMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        store: SessionStore,
        lifetime: typing.Union[int, datetime.timedelta] = 0,  # session-only
        rolling: bool = False,
        cookie_name: str = "session",
        cookie_same_site: str = "lax",
        cookie_https_only: bool = True,
        cookie_domain: typing.Optional[str] = None,
        cookie_path: typing.Optional[str] = None,
        serializer: typing.Optional[Serializer] = None,
    ) -> None:
        lifetime = int(lifetime.total_seconds() if isinstance(lifetime, datetime.timedelta) else lifetime)
        assert lifetime >= 0, "Session lifetime cannot be less than zero seconds."

        self.app = app
        self.store = store
        self.rolling = rolling
        self.serializer = serializer or JsonSerializer()
        self.cookie_name = cookie_name
        self.lifetime = lifetime
        self.cookie_domain = cookie_domain
        self.cookie_path = cookie_path
        self.security_flags = "httponly; samesite=" + cookie_same_site
        if cookie_https_only:  # Secure flag can be used with HTTPS only
            self.security_flags += "; secure"

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):  # pragma: no cover
            await self.app(scope, receive, send)
            return

        connection = HTTPConnection(scope)
        session_id = connection.cookies.get(self.cookie_name)
        handler = SessionHandler(connection, session_id, self.store, self.serializer, self.lifetime)

        scope["session"] = LoadGuard()
        scope["session_handler"] = handler

        async def send_wrapper(message: Message) -> None:
            if message["type"] != "http.response.start":
                await send(message)
                return

            # session was not loaded, do nothing
            if not handler.is_loaded:  # pragma: nocover
                await send(message)
                return

            nonlocal session_id
            path = self.cookie_path or scope.get("root_path", "") or "/"

            if handler.is_empty:
                # if session was initially empty then do nothing
                if handler.initially_empty:
                    await send(message)
                    return

                # session data loaded but empty, no matter whether it was initially empty or cleared
                # we have to remove the cookie and clear the storage
                if not self.cookie_path or self.cookie_path and scope["path"].startswith(self.cookie_path):
                    headers = MutableHeaders(scope=message)
                    header_parts = [
                        f"{self.cookie_name}=''",
                        f"path={path}",
                        "expires=Thu, 01 Jan 1970 00:00:00 GMT",
                        self.security_flags,
                    ]
                    if self.cookie_domain:
                        header_parts.append(f"domain={self.cookie_domain}")
                    header_value = "; ".join(header_parts)
                    headers.append("Set-Cookie", header_value)
                    await handler.destroy()
                await send(message)
                return

            # calculate cookie/storage expiry seconds based on selected strategy
            remaining_time = 0

            # if lifetime is zero then don't send max-age at all
            # this will create session-only cookie
            if self.lifetime > 0:
                if self.rolling:
                    # rolling strategy always extends cookie max-age by lifetime
                    remaining_time = self.lifetime
                else:
                    # non-rolling strategy reuses initial expiration date
                    remaining_time = get_session_remaining_seconds(connection)

            # persist session data
            session_id = await handler.save(remaining_time)

            headers = MutableHeaders(scope=message)
            header_parts = [
                f"{self.cookie_name}={session_id}",
                f"path={path}",
            ]

            if self.lifetime > 0:  # always send max-age for non-session scoped cookie
                header_parts.append(f"max-age={remaining_time}")

            if self.cookie_domain:
                header_parts.append(f"domain={self.cookie_domain}")

            header_parts.append(self.security_flags)
            header_value = "; ".join(header_parts)
            headers.append("set-cookie", header_value)

            await send(message)

        await self.app(scope, receive, send_wrapper)


class SessionAutoloadMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        paths: typing.Optional[typing.List[typing.Union[str, typing.Pattern[str]]]] = None,
    ) -> None:
        self.app = app
        self.paths = paths or []

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):  # pragma: no cover
            await self.app(scope, receive, send)
            return

        connection = HTTPConnection(scope, receive)
        if self.should_autoload(connection):
            await load_session(connection)

        await self.app(scope, receive, send)

    def should_autoload(self, connection: HTTPConnection) -> bool:
        if not self.paths:
            return True

        for path in self.paths:
            if re.match(path, connection.url.path):
                return True
        return False
